'''
Modified from SparseConvNet data preparation: https://github.com/facebookresearch/SparseConvNet/blob/master/examples/ScanNet/prepare_data.py
'''
import os
import open3d as o3d
import glob, plyfile, numpy as np, multiprocessing as mp, torch, json, argparse

import scannet_util
from normal_similarity_cut import NormalSegLabelGenerator

# Map relevant classes to {0,1,...,19}, and ignored classes to -100
remapper = np.ones(150) * (-100)
for i, x in enumerate([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 14, 16, 24, 28, 33, 34, 36, 39]):
    remapper[x] = i

parser = argparse.ArgumentParser()
parser.add_argument('--data_split', help='data split (train / val / test)', default='train')
parser.add_argument('--calc_normal', action="store_true", help='Calculate surface normal')
parser.add_argument('--calc_color_seg_label', action="store_true",
                    help='Segment mesh into subgraphs according to color consistency')
parser.add_argument('--calc_normal_seg_label', action="store_true",
                    help='Segment mesh into subgraphs according to surface normal consistency and record the relevant subgraph idx as label')


opt = parser.parse_args()

split = opt.data_split
print('data split: {}'.format(split))
files = sorted(glob.glob(split + '/*_vh_clean_2.ply'))
if opt.data_split != 'test':
    files2 = sorted(glob.glob(split + '/*_vh_clean_2.labels.ply'))
    files3 = sorted(glob.glob(split + '/*_vh_clean_2.0.010000.segs.json'))
    files4 = sorted(glob.glob(split + '/*[0-9].aggregation.json'))
    assert len(files) == len(files2)
    assert len(files) == len(files3)
    assert len(files) == len(files4), "{} {}".format(len(files), len(files4))

if opt.calc_normal_seg_label:
    normal_segmentor = NormalSegLabelGenerator()

def f_test(fn):
    print(fn)

    f = plyfile.PlyData().read(fn)
    points = np.array([list(x) for x in f.elements[0]])
    coords = np.ascontiguousarray(points[:, :3] - points[:, :3].mean(0))
    colors = np.ascontiguousarray(points[:, 3:6]) / 127.5 - 1

    # compute normals
    if opt.calc_normal:
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(coords)
        pcd.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30))
        normals = np.asarray(pcd.normals)

    fn3 = fn.replace('.ply', '.0.010000.segs.json')
    normal_seg_labels = normal_segmentor.get_seg_label(fn3)

    torch.save((coords, colors, normal_seg_labels), fn[:-15] + '_inst_nostuff.pth')
    print('Saving to ' + fn[:-15] + '_inst_nostuff.pth')


def f(fn):
    fn2 = fn[:-3] + 'labels.ply'
    fn3 = fn[:-15] + '_vh_clean_2.0.010000.segs.json'
    fn4 = fn[:-15] + '.aggregation.json'
    print(fn)

    f = plyfile.PlyData().read(fn)
    points = np.array([list(x) for x in f.elements[0]])
    coords = np.ascontiguousarray(points[:, :3] - points[:, :3].mean(0))
    colors = np.ascontiguousarray(points[:, 3:6]) / 127.5 - 1

    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(coords)
    if opt.calc_normal:
        pcd.estimate_normals(search_param=o3d.geometry.KDTreeSearchParamHybrid(radius=0.1, max_nn=30))
        normals = np.asarray(pcd.normals)

    if opt.calc_normal_seg_label:
        normal_seg_labels = normal_segmentor.get_seg_label(fn3)

    if opt.calc_color_seg_label:
        color_seg_labels = np.zeros_like(normal_seg_labels).fill(-1)  # we do not use color to generate implicit part

    f2 = plyfile.PlyData().read(fn2)
    sem_labels = remapper[np.array(f2.elements[0]['label'])]

    with open(fn3) as jsondata:
        d = json.load(jsondata)
        seg = d['segIndices']
    segid_to_pointid = {}
    for i in range(len(seg)):
        if seg[i] not in segid_to_pointid:
            segid_to_pointid[seg[i]] = []
        segid_to_pointid[seg[i]].append(i)

    instance_segids = []
    labels = []
    with open(fn4) as jsondata:
        d = json.load(jsondata)
        for x in d['segGroups']:
            if scannet_util.g_raw2scannetv2[x['label']] != 'wall' and scannet_util.g_raw2scannetv2[
                x['label']] != 'floor':
                instance_segids.append(x['segments'])
                labels.append(x['label'])
                assert (x['label'] in scannet_util.g_raw2scannetv2.keys())
    if (os.path.basename(fn) == 'scene0217_00_vh_clean_2.ply' and instance_segids[0] == instance_segids[
        int(len(instance_segids) / 2)]):
        instance_segids = instance_segids[: int(len(instance_segids) / 2)]
    check = []
    for i in range(len(instance_segids)): check += instance_segids[i]
    assert len(np.unique(check)) == len(check)

    instance_labels = np.ones(sem_labels.shape[0]) * -100
    for i in range(len(instance_segids)):
        segids = instance_segids[i]
        pointids = []
        for segid in segids:
            pointids += segid_to_pointid[segid]
        instance_labels[pointids] = i
        assert (len(np.unique(sem_labels[pointids])) == 1)

    if not opt.calc_normal:
        torch.save((coords, colors, sem_labels, instance_labels),
                   fn[:-15] + '_inst_nostuff.pth')
    else:
        torch.save((coords, colors, sem_labels, instance_labels, normals, color_seg_labels, normal_seg_labels),
                   fn[:-15] + '_inst_nostuff.pth')

    print('Saving to ' + fn[:-15] + '_inst_nostuff.pth')

if __name__ == '__main__':
    p = mp.Pool(processes=mp.cpu_count())
    if opt.data_split == 'test':
        p.map(f_test, files)
    else:
        p.map(f, files)
    p.close()
    p.join()
